# -*- coding: utf-8 -*-

# Copyright 2017, IBM.
#
# This source code is licensed under the Apache License, Version 2.0 found in
# the LICENSE.txt file in the root directory of this source tree.

"""C++ qasm_simulator gates."""
from .load import load
from .noise import noise
from .save import save
from .snapshot import snapshot
from .wait import wait
